// HeadtrackerDlg.cpp : implementation file
//
#include "stdafx.h"
// #include "string.h"
#include "Headtracker.h"
#include "HeadtrackerDlg.h"
#include <sys/timeb.h>	// time()
#include <time.h>		// time()

#include "digitizer.h"
#include "track.h"
#include "base.h" // ISIZE[XY]
#include "camera.h"
#include "color_histogram.h"	// BGRToColorBins

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define LEFT_MARGIN 10
#define TOP_MARGIN 10
#define MAX_STRING_LENGTH 80
#define DIALOG_BACKGROUND_COLOR 192 // This is the dialog background color
#define RECORDING_DOT_X 236		// location of red dot indicating recording
#define RECORDING_DOT_Y 445
#define TRACKING_DOT_X 50		// location of red dot indicating tracker locked
#define TRACKING_DOT_Y 373
#define TIME_DOT_X 595			// location of red dot indicating sequence is being timed
#define TIME_DOT_Y 473
#define TIMER_ID	1
#define BYTES_PER_PIXEL_IN_SEQUENCE 3 // keep this at 3 for now
#define N_DISPLAY_SPEEDS 8
#define INITIAL_DISPLAY_SPEED 4

Digitizer_Params dparams;	
HANDLE imageready;
HANDLE VideoThread;
CDC *my_cdc;
HDC my_hdc;
enum {LIVE_VIDEO, IMAGE_SEQUENCE};
enum {DIG_BUSTER = 0, DIG_WINNOV = 1};
BOOL recording_sequence = FALSE;
enum {PLAY, STOP, FORWARD, BACKWARD};
int playmode = STOP;
int numfiles;
CHeadtrackerDlg *headtrackerDlg;
ImageBGR24 Image(ISIZEX, ISIZEY);
ImageBGR24 img_reddot(8,8);
Image8 img_blankdot(8,8);
char fname[MAX_STRING_LENGTH];
char filename_model[] = "model_stanb.dat";

extern Image8 img_gray;
extern Image16 img_gauss;
extern Image8 img_grad;
extern Image16 img_gradx;
extern Image16 img_grady;
extern Image8 img_gradx8;
extern Image8 img_grady8;
extern Image16 img_tmp16;
extern ImageBGR24 img_outline;
extern ImageBGR24 img_outline_grad;
extern ImageBGR24 img_outline_color;
extern Image8 img_outline8;
extern Image8 img_outline_grad8;
extern Image8 img_outline_color8;
extern Image8 img_mask;
extern Image8 img_color1;
extern Image8 img_color2;
extern Image8 img_color3;
extern Image8 img_color1ref;
extern Image8 img_color2ref;
extern Image8 img_color3ref;
extern Image8 img_color1q;
extern Image8 img_color2q;
extern Image8 img_color3q;
extern Image8 img_likelihood;
extern Image8 img_backprojection;
extern Image8 img_lights;
extern Image8 img_bin;
static Image8 img_tmp8(ISIZEX,ISIZEY);
static ImageBGR24 img_tmp24(ISIZEX,ISIZEY);
//extern Image8 img_ch_model;
//extern Image8 img_ch_new;
extern int tmp1, tmp2;
extern int tmp3;
Image8 blankimg(ISIZEX, 256);
BOOL building_model = FALSE;
static char msg[80];
BOOL use_gradient, use_color; // evidence used by tracker
BOOL sum_gradient_magnitude;
BOOL zoom_control;
BOOL adapt_histogram;
BOOL show_details;
unsigned char *recorded_images;
int *ellipse_states;
EllipseState curr_state;	// The current state (x,y,sz) of the ellipse
char fname_ellipse_states[] = "ellipse_states.txt";
static int display_speeds[N_DISPLAY_SPEEDS] = {1, 6, 10, 15, 30, 60, 90, 1000};
static int display_speed;

/////////////////////////////////////////////////////////////////////////////
// CAboutDlg dialog used for App About

class CAboutDlg : public CDialog
{
public:
	CAboutDlg();

// Dialog Data
	//{{AFX_DATA(CAboutDlg)
	enum { IDD = IDD_ABOUTBOX };
	//}}AFX_DATA

	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CAboutDlg)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

// Implementation
protected:
	//{{AFX_MSG(CAboutDlg)
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

CAboutDlg::CAboutDlg() : CDialog(CAboutDlg::IDD)
{
	//{{AFX_DATA_INIT(CAboutDlg)
	//}}AFX_DATA_INIT
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CAboutDlg)
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CAboutDlg, CDialog)
	//{{AFX_MSG_MAP(CAboutDlg)
		// No message handlers
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CHeadtrackerDlg dialog

CHeadtrackerDlg::CHeadtrackerDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CHeadtrackerDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CHeadtrackerDlg)
	m_input_mode = -1;
	m_enable_recording = FALSE;
	m_last_filenum = 0;
	m_basename = _T("");
	m_first_filenum = 0;
	m_tracking = FALSE;
	m_building_model = FALSE;
	m_use_camera = -1;
	m_input_file_bw_instead_of_color = -1;
	m_filenum = 0;
	//}}AFX_DATA_INIT
	// Note that LoadIcon does not require a subsequent DestroyIcon in Win32
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);
}

void CHeadtrackerDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CHeadtrackerDlg)
	DDX_CBIndex(pDX, IDC_INPUT_MODE, m_input_mode);
	DDX_Check(pDX, IDC_ENABLE_RECORDING, m_enable_recording);
	DDX_Text(pDX, IDC_LAST, m_last_filenum);
	DDV_MinMaxInt(pDX, m_last_filenum, 0, 9999);
	DDX_Text(pDX, IDC_BASENAME, m_basename);
	DDX_Text(pDX, IDC_FIRST, m_first_filenum);
	DDV_MinMaxInt(pDX, m_first_filenum, 0, 9999);
	DDX_Check(pDX, IDC_CHECK_TRACKING, m_tracking);
	DDX_Check(pDX, IDC_BUILDING_MODEL, m_building_model);
	DDX_Radio(pDX, IDC_CAMERA_OFF, m_use_camera);
	DDX_Radio(pDX, IDC_INPUT_FILE_COLOR, m_input_file_bw_instead_of_color);
	DDX_Text(pDX, IDC_FRAME_NUMBER, m_filenum);
	DDV_MinMaxInt(pDX, m_filenum, 0, 9999);
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CHeadtrackerDlg, CDialog)
	//{{AFX_MSG_MAP(CHeadtrackerDlg)
	ON_WM_SYSCOMMAND()
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	ON_WM_DESTROY()
	ON_CBN_SELCHANGE(IDC_INPUT_MODE, OnSelchangeInputMode)
	ON_BN_CLICKED(IDC_CHECK_TRACKING, OnCheckTracking)
	ON_BN_CLICKED(IDC_RECORD_SEQ, OnRecordSeq)
	ON_WM_TIMER()
	ON_BN_CLICKED(IDC_PLAY_SEQ, OnPlaySeq)
	ON_BN_CLICKED(IDC_FORWARD, OnForward)
	ON_BN_CLICKED(IDC_BACKWARD, OnBackward)
	ON_BN_CLICKED(IDC_STOP_PLAYING, OnStopPlaying)
	ON_EN_CHANGE(IDC_BASENAME, OnChangeBasename)
	ON_WM_LBUTTONDOWN()
	ON_BN_CLICKED(IDC_SAVE_MODEL, OnSaveModel)
	ON_BN_CLICKED(IDC_BUILDING_MODEL, OnBuildingModel)
	ON_BN_CLICKED(IDC_SHOW_PIXEL_LISTS, OnShowPixelLists)
	ON_BN_CLICKED(IDC_CAMERA_HOME, OnCameraHome)
	ON_BN_CLICKED(IDC_LOCK_WHITE_BALANCE, OnLockWhiteBalance)
	ON_BN_CLICKED(IDC_USE_GRADIENT, OnUseGradient)
	ON_BN_CLICKED(IDC_USE_COLOR, OnUseColor)
	ON_BN_CLICKED(IDC_ZOOM, OnZoom)
	ON_BN_CLICKED(IDC_ADAPT_HISTOGRAM, OnAdaptHistogram)
	ON_BN_CLICKED(IDC_CAMERA_OFF, OnCameraOff)
	ON_BN_CLICKED(IDC_CAMERA_ON, OnCameraOn)
	ON_BN_CLICKED(IDC_GRADIENT_MAG, OnGradientMag)
	ON_BN_CLICKED(IDC_GRADIENT_DOT, OnGradientDot)
	ON_BN_CLICKED(IDC_TIME_SEQUENCE, OnTimeSequence)
	ON_BN_CLICKED(IDC_SAVE_ERROR_FUNCTION, OnSaveErrorFunction)
	ON_BN_CLICKED(IDC_OPEN_MANUAL, OnOpenManual)
	ON_BN_CLICKED(IDC_CLOSE_MANUAL, OnCloseManual)
	ON_BN_CLICKED(IDC_SHOW_DETAILS, OnShowDetails)
	ON_BN_CLICKED(IDC_OVERLAY_PRECOMPUTED_ELLIPSES, OnOverlayPrecomputedEllipses)
	ON_BN_CLICKED(IDC_CONTINUE_PLAYING, OnContinuePlaying)
	ON_WM_HSCROLL()
	ON_BN_CLICKED(IDC_INPUT_FILE_BW, OnInputFileBw)
	ON_BN_CLICKED(IDC_INPUT_FILE_COLOR, OnInputFileColor)
	ON_EN_KILLFOCUS(IDC_LAST, OnKillfocusLast)
	ON_EN_KILLFOCUS(IDC_FIRST, OnKillfocusFirst)
	ON_EN_KILLFOCUS(IDC_BASENAME, OnKillfocusBasename)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


/////////////////////////////////////////////////////////////////////////////
// 

void PrintDisplaySpeed(int speed)
{
	sprintf(msg, (speed<0 ?		"                         " : 
								"Speed:  %4d Hz  "), speed);
	my_cdc->TextOut(465, 520, msg);
}

void PrintEditNumber4D(int id, int num)
{
	sprintf(msg, "%4d", num);
	headtrackerDlg->GetDlgItem(id)->SetWindowText(msg);
}

void PrintEditString(int id, char *str)
{
	headtrackerDlg->GetDlgItem(id)->SetWindowText(str);
}

void CHeadtrackerDlg::EnableDisableWindows() 
{
	switch (m_input_mode) {
	case LIVE_VIDEO:
		GetDlgItem(IDC_STATIC_CAMERA)->EnableWindow(TRUE);
		GetDlgItem(IDC_CAMERA_ON)->EnableWindow(TRUE);
		GetDlgItem(IDC_CAMERA_OFF)->EnableWindow(TRUE);
		GetDlgItem(IDC_ZOOM)->EnableWindow(TRUE);
		GetDlgItem(IDC_CAMERA_HOME)->EnableWindow(TRUE);
		GetDlgItem(IDC_LOCK_WHITE_BALANCE)->EnableWindow(TRUE);
		GetDlgItem(IDC_RECORD_SEQ)->EnableWindow(TRUE);
		GetDlgItem(IDC_ENABLE_RECORDING)->EnableWindow(TRUE);
		GetDlgItem(IDC_INPUT_FILE_COLOR)->EnableWindow(FALSE);
		GetDlgItem(IDC_INPUT_FILE_BW)->EnableWindow(FALSE);
		GetDlgItem(IDC_PLAY_SEQ)->EnableWindow(FALSE);
		GetDlgItem(IDC_FORWARD)->EnableWindow(FALSE);
		GetDlgItem(IDC_BACKWARD)->EnableWindow(FALSE);
		GetDlgItem(IDC_STOP_PLAYING)->EnableWindow(TRUE);
		GetDlgItem(IDC_STATIC_FRAME_NUMBER)->EnableWindow(FALSE);
		GetDlgItem(IDC_FRAME_NUMBER)->EnableWindow(FALSE);
		GetDlgItem(IDC_SCROLLBAR_SPEED)->EnableWindow(FALSE);
		GetDlgItem(IDC_TIME_SEQUENCE)->EnableWindow(FALSE);
		GetDlgItem(IDC_OVERLAY_PRECOMPUTED_ELLIPSES)->EnableWindow(FALSE);
		PrintDisplaySpeed(-1);
		break;
	case IMAGE_SEQUENCE:
		GetDlgItem(IDC_STATIC_CAMERA)->EnableWindow(FALSE);
		GetDlgItem(IDC_CAMERA_ON)->EnableWindow(FALSE);
		GetDlgItem(IDC_CAMERA_OFF)->EnableWindow(FALSE);
		GetDlgItem(IDC_ZOOM)->EnableWindow(FALSE);
		GetDlgItem(IDC_CAMERA_HOME)->EnableWindow(FALSE);
		GetDlgItem(IDC_LOCK_WHITE_BALANCE)->EnableWindow(FALSE);
		GetDlgItem(IDC_RECORD_SEQ)->EnableWindow(FALSE);
		GetDlgItem(IDC_ENABLE_RECORDING)->EnableWindow(FALSE);
		GetDlgItem(IDC_INPUT_FILE_COLOR)->EnableWindow(TRUE);
		GetDlgItem(IDC_INPUT_FILE_BW)->EnableWindow(TRUE);
		GetDlgItem(IDC_PLAY_SEQ)->EnableWindow(TRUE);
		GetDlgItem(IDC_FORWARD)->EnableWindow(TRUE);
		GetDlgItem(IDC_BACKWARD)->EnableWindow(TRUE);
		GetDlgItem(IDC_STOP_PLAYING)->EnableWindow(TRUE);
		GetDlgItem(IDC_STATIC_FRAME_NUMBER)->EnableWindow(TRUE);
		GetDlgItem(IDC_FRAME_NUMBER)->EnableWindow(TRUE);
		GetDlgItem(IDC_SCROLLBAR_SPEED)->EnableWindow(TRUE);
		GetDlgItem(IDC_TIME_SEQUENCE)->EnableWindow(TRUE);
		GetDlgItem(IDC_OVERLAY_PRECOMPUTED_ELLIPSES)->EnableWindow(TRUE);
		PrintDisplaySpeed(display_speed);
		break;
	}

}

void CHeadtrackerDlg::EnableDisableMagDot(BOOL ug) 
{
	GetDlgItem(IDC_GRADIENT_MAG)->EnableWindow(ug);
	GetDlgItem(IDC_GRADIENT_DOT)->EnableWindow(ug);
}

void TrackAndDisplay()
{
	static BOOL tracker_locked_old = FALSE;
	int left = 1, top = 0;

//	Digitizer_Shutdown(0);
	track(&Image, &curr_state);
	if (headtrackerDlg->m_use_camera) MoveCamera(&curr_state);
//	track(&Image, use_camera);
//	sprintf(msg, "%3d %3d", tmp1, tmp2);
//	headtrackerDlg->GetDlgItem(IDC_XY)->SetWindowText(msg);
//	sprintf(msg, "%d", tmp3);
//	headtrackerDlg->GetDlgItem(IDC_BGR)->SetWindowText(msg);

//		img_reddot.Display(my_hdc,TRACKING_DOT_X, TRACKING_DOT_Y);
//		img_gray.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
//		img_gauss.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
		img_outline.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
		if (show_details) {
			img_outline_grad8.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
			img_outline_color8.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
			img_grad.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
	//		img_bspline2.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
			left = 0; top++;
	//		img_tmp16.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
	//		img_gradx.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
	//		img_grady.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
			img_likelihood.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
			img_bin.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
			img_backprojection.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
			img_gradx8.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
			img_grady8.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
	//		img_lights.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
	//		img_mask.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
	//		img_color1q.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
	//		img_color2q.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
	//		img_color3q.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
	//		img_ch_model.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
	//		img_ch_new.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
	//		img_overlay.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
	//		img_overlay2.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
		}

//	sprintf(msg, "camera:  %4d        ", headtrackerDlg->m_use_camera);
//	sprintf(msg, "camera:  %4d        ", use_camera);
//	my_cdc->TextOut(5, 215, msg);
}

static int VideoLoop(int reserved)
{
	Image8 tmpimg(ISIZEX,ISIZEY);	
	DWORD timestamp;
//	int left, top;

	while (1)
	{
		if (WaitForSingleObject(imageready,3000) != WAIT_FAILED) //wait for framegrabber pulse
		{
			if (headtrackerDlg->m_input_mode == LIVE_VIDEO)  {
//				left = 0; top = 0;
				if (Digitizer_FillBuffer(0,&Image,&timestamp)) {
					Image.Display(my_hdc,LEFT_MARGIN,TOP_MARGIN);
				}
				if (headtrackerDlg->m_tracking)  {
					TrackAndDisplay();
				}
				if (recording_sequence)  {
					int filenum_rel = headtrackerDlg->m_filenum - headtrackerDlg->m_first_filenum;
					int *ptres = ellipse_states+filenum_rel*3;
					memcpy(recorded_images+(filenum_rel)*BYTES_PER_PIXEL_IN_SEQUENCE*ISIZEX*ISIZEY,
						Image.GetImagePtr(), BYTES_PER_PIXEL_IN_SEQUENCE*ISIZEX*ISIZEY);
//						sprintf(msg, "fn %3d state %3d %3d %3d       ", filenum_rel, curr_state.x, curr_state.y, curr_state.sz);
//						my_cdc->TextOut(5,220,msg);
					*ptres++ = curr_state.x;
					*ptres++ = curr_state.y;
					*ptres++ = curr_state.sz;
					sprintf(msg, "%4d", headtrackerDlg->m_filenum);
					headtrackerDlg->GetDlgItem(IDC_FRAME_NUMBER)->SetWindowText(msg);
					headtrackerDlg->m_filenum++;
//					sprintf(msg, "m_filenum %d < %d relative %d            ", m_filenum, headtrackerDlg->m_last_filenum,
//						m_filenum-headtrackerDlg->m_first_filenum);
//					my_cdc->TextOut(5,220,msg);
					if (headtrackerDlg->m_filenum > headtrackerDlg->m_last_filenum) {
//							Digitizer_Shutdown(0);
						int i, i_rel;
						recording_sequence = FALSE;
//						sprintf(msg, "                           ", m_filenum, headtrackerDlg->m_last_filenum);
//						my_cdc->TextOut(5,220,msg);
						sprintf(fname, fname_ellipse_states);
						FILE *fp = fopen(fname, "w");
						if (fp==NULL) {
							sprintf(msg, "Couldn't open file '%s' for writing!!", fname);
							AfxMessageBox(msg, MB_OK|MB_ICONSTOP, 0);
							Digitizer_Shutdown(0);
						}
						fprintf(fp, "frame x   y sz\n");
						for (i=headtrackerDlg->m_first_filenum ; i<=headtrackerDlg->m_last_filenum ; i++) {
							i_rel = i - headtrackerDlg->m_first_filenum;
							sprintf(fname, headtrackerDlg->m_basename, i_rel);
							if (BYTES_PER_PIXEL_IN_SEQUENCE==1) {
								memcpy(img_tmp8.GetImagePtr(), 
									recorded_images+i_rel*BYTES_PER_PIXEL_IN_SEQUENCE*ISIZEX*ISIZEY, BYTES_PER_PIXEL_IN_SEQUENCE*ISIZEX*ISIZEY);
								img_tmp8.Write8BitBMP(fname);
							} else if (BYTES_PER_PIXEL_IN_SEQUENCE==3) {
								memcpy(img_tmp24.GetImagePtr(), 
									recorded_images+i_rel*BYTES_PER_PIXEL_IN_SEQUENCE*ISIZEX*ISIZEY, BYTES_PER_PIXEL_IN_SEQUENCE*ISIZEX*ISIZEY);
								img_tmp24.Write24BitBMP(fname);
							}
							fprintf(fp, "%3d %3d %3d %2d\n", i, 
								*(ellipse_states+3*i_rel), *(ellipse_states+3*i_rel+1), *(ellipse_states+3*i_rel+2));
						}
						fclose(fp);

//						fp=fopen("tmp.dat", "wb");
//						sprintf(msg, "file %d   ", fp);
//						my_cdc->TextOut(5,220,msg);
//						fwrite(ellipse_states, sizeof(int), 51, fp);
//						fclose(fp);
						free(recorded_images);
						free(ellipse_states);
						headtrackerDlg->m_use_camera = FALSE;
						StopCamera();
						((CButton *) headtrackerDlg->GetDlgItem(IDC_CAMERA_OFF))->SetCheck(1);
						((CComboBox *) headtrackerDlg->GetDlgItem(IDC_INPUT_MODE))->SetCurSel(IMAGE_SEQUENCE);
						headtrackerDlg->m_input_mode = IMAGE_SEQUENCE;
						headtrackerDlg->EnableDisableWindows();
						img_blankdot.Display(my_hdc,RECORDING_DOT_X, RECORDING_DOT_Y);
					}
				}
			}
		}
		else 
			;//?
	}
	
	return 0;
}

void StartVideoThread(void)
{
	static DWORD threadIdentifier;
	VideoThread = CreateThread( NULL,	// use default security
		 						0,		// use default stack size
						  		(LPTHREAD_START_ROUTINE) &VideoLoop,
						  		0,		// thread parameter value
						  		CREATE_SUSPENDED,
						  		&threadIdentifier );
	SetThreadPriority ( VideoThread, 
						THREAD_PRIORITY_TIME_CRITICAL );		
	ResumeThread ( VideoThread );
}

/////////////////////////////////////////////////////////////////////////////
// CHeadtrackerDlg message handlers

BOOL CHeadtrackerDlg::OnInitDialog()
{
	m_input_mode = LIVE_VIDEO;
//	m_input_mode = IMAGE_SEQUENCE;
	m_use_camera = FALSE;
//	m_basename = "../../seq1/img%03d.bmp";
	m_basename = "../../seq_stans_head/img%03d.bmp";
//	m_basename = "../../seq_ptz_complex2/img%03d.bmp";
//	m_basename = "../../seq_head1/img%03d.bmp";
//	m_basename = "../../seq_untex/untex_%04d.bmp";
//	m_basename = "../../seq_untexb_largehead/untexb_%04d.bmp";
	m_first_filenum = 0;
	m_last_filenum = 500;
	m_filenum = m_first_filenum;
	m_enable_recording = FALSE;
	m_tracking = FALSE;
//	m_tracking = TRUE;
	use_gradient = TRUE;
	use_color = TRUE;
	zoom_control = FALSE;
	adapt_histogram = FALSE;
	sum_gradient_magnitude = FALSE;
	show_details = FALSE;
	m_input_file_bw_instead_of_color = FALSE;
	display_speed = display_speeds[INITIAL_DISPLAY_SPEED];

	ReadModel(filename_model);
	CDialog::OnInitDialog();

	{
		CScrollBar *pSB = (CScrollBar *) GetDlgItem(IDC_SCROLLBAR_SPEED);
		pSB->SetScrollRange(0, N_DISPLAY_SPEEDS-1);
		pSB->SetScrollPos(INITIAL_DISPLAY_SPEED, TRUE);
	}

	((CButton *) GetDlgItem(IDC_USE_GRADIENT))->SetCheck(use_gradient);
	((CButton *) GetDlgItem(IDC_USE_COLOR))->SetCheck(use_color);
	((CButton *) GetDlgItem(IDC_ZOOM))->SetCheck(zoom_control);
	((CButton *) GetDlgItem(IDC_ADAPT_HISTOGRAM))->SetCheck(adapt_histogram);
	((CButton *) GetDlgItem(IDC_GRADIENT_MAG))->SetCheck(sum_gradient_magnitude);
	((CButton *) GetDlgItem(IDC_GRADIENT_DOT))->SetCheck(1-sum_gradient_magnitude);
	((CButton *) GetDlgItem(IDC_SHOW_DETAILS))->SetCheck(show_details);

	my_cdc = GetDC();
	my_hdc = GetDC()->m_hDC;

	EnableDisableWindows();
	headtrackerDlg = (CHeadtrackerDlg *) (AfxGetApp()->m_pMainWnd);
	TRACE("here %d\n", headtrackerDlg);

	///// ->
	// Add "About..." menu item to system menu.

	// IDM_ABOUTBOX must be in the system command range.
	ASSERT((IDM_ABOUTBOX & 0xFFF0) == IDM_ABOUTBOX);
	ASSERT(IDM_ABOUTBOX < 0xF000);

	CMenu* pSysMenu = GetSystemMenu(FALSE);
	CString strAboutMenu;
	strAboutMenu.LoadString(IDS_ABOUTBOX);
	if (!strAboutMenu.IsEmpty())
	{
		pSysMenu->AppendMenu(MF_SEPARATOR);
		pSysMenu->AppendMenu(MF_STRING, IDM_ABOUTBOX, strAboutMenu);
	}

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon
	///// <-

	//	CMyCheckListBox mychecklist;
	// To change from SVIDEO to NTSC, manually edit the file C:\PROGRAM FILES\WINNOV VIDEUM\VIDEUM.INI: Change 2001 to 4001 for FmtFlagsLo
	// dparams.NTSC = TRUE;			// Note this is ignored in VFWs case
	// dparams.SVIDEO = 0;			// Note this is ignored in VFWs case	
	dparams.HALFRES = TRUE;
	dparams.sizex = ISIZEX;
	dparams.sizey = ISIZEY;
	dparams.EveryNthFrame = 1;
	dparams.TotalSize = ISIZEX*ISIZEY;
	dparams.ImageFormatDesired = APTOS_BGR24PACKED;// YUV_24;
	dparams.CaptureFormatDesired = APTOS_BGR24PACKED;
	dparams.app_win = this->m_hWnd;
	dparams.Device_Number = DIG_WINNOV; // 0 -> Buster, 1 - > Winnov

	imageready=CreateEvent(NULL,FALSE,FALSE,NULL);
	dparams.event = imageready;

	if (!OpenCamera(2)) {
		AfxMessageBox("Failed to Open Camera", MB_OK|MB_ICONSTOP, 0);
	}

	InitTracker(&curr_state);

	if (Digitizer_Init(0,dparams) == TRUE) {
		StartVideoThread();
	} else
		AfxMessageBox("Failed to Connect To Digitizer",MB_OK|MB_ICONSTOP,0);

	blankimg = DIALOG_BACKGROUND_COLOR;
	img_reddot.Set(0,0,255);
	img_blankdot = DIALOG_BACKGROUND_COLOR;

	if (SetTimer(TIMER_ID, 1000/display_speeds[INITIAL_DISPLAY_SPEED], NULL) == 0)
		AfxMessageBox("Failed to Set Timer",MB_OK|MB_ICONSTOP,0);

	my_cdc->SetBkColor(0x00C0C0C0);
	PrintEditNumber4D(IDC_FRAME_NUMBER, m_filenum);

	return TRUE;  // return TRUE  unless you set the focus to a control
}

// Without the OnOK function, pressing the RETURN key inside an edit box
// causes the dialog to exit.
void CHeadtrackerDlg::OnOK() 
{
	int edit_box_id = GetFocus()->GetDlgCtrlID();
	UpdateData(TRUE);

	switch (edit_box_id) {
	case IDC_FRAME_NUMBER: 
		if (m_filenum < m_first_filenum || m_filenum > m_last_filenum) m_filenum = m_first_filenum;
		PrintEditNumber4D(IDC_FRAME_NUMBER, m_filenum);
		break;
	case IDC_BASENAME:
		PrintEditString(IDC_BASENAME, (char *) ((LPCTSTR) m_basename));
		break;
	case IDC_FIRST:
		PrintEditNumber4D(IDC_FIRST, m_first_filenum);
		if (m_filenum < m_first_filenum) m_filenum = m_first_filenum;
		break;
	case IDC_LAST:
		PrintEditNumber4D(IDC_LAST, m_last_filenum);
		if (m_filenum > m_last_filenum) m_filenum = m_last_filenum;
		break;
	default:
		break;
	}
}

void CHeadtrackerDlg::OnSysCommand(UINT nID, LPARAM lParam)
{
	if ((nID & 0xFFF0) == IDM_ABOUTBOX)
	{
		CAboutDlg dlgAbout;
		dlgAbout.DoModal();
	}
	else
	{
		CDialog::OnSysCommand(nID, lParam);
	}
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.

void CHeadtrackerDlg::OnPaint() 
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, (WPARAM) dc.GetSafeHdc(), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CDialog::OnPaint();
		if (m_input_mode == IMAGE_SEQUENCE) PrintDisplaySpeed(display_speed);
	}
}

// The system calls this to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CHeadtrackerDlg::OnQueryDragIcon()
{
	return (HCURSOR) m_hIcon;
}

void CHeadtrackerDlg::OnDestroy() 
{
	CDialog::OnDestroy();
	
	// TODO: Add your message handler code here
	Sleep(2000);
//	Digitizer_Stop(0);
	Digitizer_Shutdown(0);
	StopCamera();
	CloseCamera();
}


/////////////////////////////////////////////////////////////////////////////
// General functions

void NegateBoolean(BOOL *b) { *b = 1 - *b; }


/////////////////////////////////////////////////////////////////////////////
// Input Mode (Live vs. Image Sequence) functions

void CHeadtrackerDlg::OnSelchangeInputMode() 
{
	UpdateData(TRUE);
	TRACE("input mode: %d\n", m_input_mode);
	EnableDisableWindows();
}


/////////////////////////////////////////////////////////////////////////////
// Tracking functions

void CHeadtrackerDlg::OnUseGradient() { 
	if (use_color) {
		NegateBoolean(&use_gradient); 
		EnableDisableMagDot(use_gradient);
	}
	else ((CButton *) GetDlgItem(IDC_USE_GRADIENT))->SetCheck(use_gradient);
}
void CHeadtrackerDlg::OnUseColor() { 
	if (use_gradient) NegateBoolean(&use_color); 
	else ((CButton *) GetDlgItem(IDC_USE_COLOR))->SetCheck(use_color);
}
void CHeadtrackerDlg::OnGradientMag() {	sum_gradient_magnitude = TRUE; }
void CHeadtrackerDlg::OnGradientDot() {	sum_gradient_magnitude = FALSE; }
void CHeadtrackerDlg::OnSaveModel() { WriteModel(filename_model); }

void CHeadtrackerDlg::OnCheckTracking() 
{
	if (!m_tracking) {	// start tracking
	//	InitTracker(&curr_state);  // redundant now
	} else {			// stop tracking
	//	if (!Image.Write24BitBMP("image.bmp")) {
	//		AfxMessageBox("Error writing BMP file",MB_OK|MB_ICONSTOP,0);
	//	}
	}
	m_tracking = 1 - m_tracking;
}

void CHeadtrackerDlg::OnAdaptHistogram() 
{
	NegateBoolean(&adapt_histogram);
	if (!adapt_histogram) CopyModelHistogramToRef();
}

void CHeadtrackerDlg::OnBuildingModel() 
{
	UpdateData(TRUE);
	building_model = m_building_model;
	if (building_model) {
		InitBuildingOfModel();
	} else {
		StopBuildingOfModel();
	}
}

void CHeadtrackerDlg::OnShowDetails() 
{
	int left = 0, top = 0;

	NegateBoolean(&show_details);
	if (!show_details) {
		blankimg.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
		blankimg.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
		blankimg.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
		blankimg.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
		blankimg.Display(my_hdc,(left+1)*LEFT_MARGIN+left*ISIZEX,(top+1)*TOP_MARGIN+top*ISIZEY); left++;
	}

}


/////////////////////////////////////////////////////////////////////////////
// Image Sequence functions

FILE *fp_ellipses = NULL;

void CHeadtrackerDlg::OnInputFileBw() { UpdateData(TRUE); }
void CHeadtrackerDlg::OnInputFileColor() { UpdateData(TRUE); }

void CHeadtrackerDlg::OnRecordSeq() 
{
//	int can_allocate_memory = TRUE;

	UpdateData(TRUE);
	TRACE("1 %d %d %d\n", m_enable_recording, m_first_filenum, m_last_filenum);
	if (m_enable_recording && m_first_filenum <= m_last_filenum) {
		TRACE("2\n");
		numfiles = m_last_filenum - m_first_filenum + 1;
		recorded_images = (unsigned char *) 
			malloc(numfiles*BYTES_PER_PIXEL_IN_SEQUENCE*ISIZEX*ISIZEY*sizeof(unsigned char));
		if (recorded_images == NULL) {
			AfxMessageBox("Unable to allocate memory for image sequence",MB_OK|MB_ICONSTOP,0);
			return;
//			can_allocate_memory = FALSE;
		}
		ellipse_states = (int *) malloc(numfiles*3*sizeof(int));
		if (ellipse_states == NULL) {
			AfxMessageBox("Unable to allocate memory for ellipse states",MB_OK|MB_ICONSTOP,0);
			return;
//			can_allocate_memory = FALSE;
		}
		m_filenum = m_first_filenum;
//		for (i = 0 ; i < numfiles ; i++)  {
//			image_sequence_Y[i] = new Image8(ISIZEX, ISIZEY);
//			image_sequence_U[i] = new Image8(ISIZEX, ISIZEY);
//			image_sequence_V[i] = new Image8(ISIZEX, ISIZEY);
//		}
//		if (can_allocate_memory) {
		img_reddot.Display(my_hdc,RECORDING_DOT_X, RECORDING_DOT_Y);
		recording_sequence = TRUE;
//		}
	}
}

void CHeadtrackerDlg::OnTimer(UINT nIDEvent) 
{
//	CDialog::OnTimer(nIDEvent);
//	char msg[80];

	if (m_input_mode == IMAGE_SEQUENCE)  {
		if (playmode != STOP)  {
			sprintf(fname, m_basename, m_filenum);
			if (!m_input_file_bw_instead_of_color && !Image.Read24BitBMP(fname)) { // color
				playmode = STOP;
				AfxMessageBox("Error reading BMP file (Perhaps image is B/W, not Color)",MB_OK|MB_ICONSTOP,0);
			} else if (m_input_file_bw_instead_of_color && !img_gray.Read8BitBMP(fname)) { // b/w
				playmode = STOP;
				AfxMessageBox("Error reading BMP file (Perhaps image is Color, not B/W)",MB_OK|MB_ICONSTOP,0);
			} else {
				unsigned char *ptr = img_gray.GetImagePtr();
				if (m_input_file_bw_instead_of_color) Image.FillWithRGB(ptr, ptr, ptr); // b/w
				Image.Display(my_hdc,LEFT_MARGIN,TOP_MARGIN);			
				PrintEditNumber4D(IDC_FRAME_NUMBER, m_filenum);
//				sprintf(msg, "%4d", m_filenum);
//				GetDlgItem(IDC_FRAME_NUMBER)->SetWindowText(msg);
				if (headtrackerDlg->m_tracking)  {
					TrackAndDisplay();
					if (fp_ellipses) 
						fprintf(fp_ellipses, "%3d %3d %3d %3d ; ...\n", 
							m_filenum, curr_state.x, curr_state.y, curr_state.sz);
//					sprintf(fname, "../../seq_results/overlay%03d.bmp", m_filenum);
//					if (!img_outline8.Write8BitBMP(fname)) {
//						AfxMessageBox("Error writing BMP file",MB_OK|MB_ICONSTOP,0);
//					}
				}
			}
		}
		if (playmode == PLAY)  {
			if (m_filenum == m_last_filenum) {
				fprintf(fp_ellipses, "];\n");
				fclose(fp_ellipses);
				playmode = STOP;
				img_reddot.Display(my_hdc,TRACKING_DOT_X, TRACKING_DOT_Y);
			}
			else m_filenum++;
		}
		else playmode = STOP;
	}
}

void CHeadtrackerDlg::OnPlaySeq() 
{
	img_blankdot.Display(my_hdc,TRACKING_DOT_X, TRACKING_DOT_Y);
	if (m_first_filenum <= m_last_filenum) {
		if (fp_ellipses != NULL) fclose(fp_ellipses);
		fp_ellipses = fopen("../../seq_results/loc.m", "w");
		if (fp_ellipses == NULL) {
			AfxMessageBox("Error opening 'loc.m' file", MB_OK|MB_ICONSTOP,0);
		} else {
			fprintf(fp_ellipses, "loc = [...\n");
			m_filenum = m_first_filenum;
			playmode = PLAY;
		}
	}
}

void CHeadtrackerDlg::OnContinuePlaying() 
{
	if (m_first_filenum <= m_last_filenum) {
		if (playmode==STOP || playmode==BACKWARD) m_filenum++;
		if (m_filenum >= m_first_filenum && m_filenum <= m_last_filenum) playmode = PLAY;
	}
}

void CHeadtrackerDlg::OnForward() 
{
	if (m_filenum < m_last_filenum)  {
		m_filenum++;
		playmode = FORWARD;
	}
}

void CHeadtrackerDlg::OnBackward() 
{
	if (m_filenum > m_first_filenum)  {
		m_filenum--;
		playmode = BACKWARD;
	}
}

void CHeadtrackerDlg::OnStopPlaying() 
{
	if (m_input_mode==LIVE_VIDEO && recording_sequence) {
		m_filenum = headtrackerDlg->m_last_filenum - 1;
	} else if (m_input_mode==IMAGE_SEQUENCE) {
		if (playmode==PLAY || playmode==FORWARD) m_filenum--;
		else if (playmode==BACKWARD) m_filenum++;
		playmode = STOP;
	}
}

void CHeadtrackerDlg::OnChangeBasename() { }//UpdateData(TRUE); }
void CHeadtrackerDlg::OnKillfocusBasename() { PrintEditString(IDC_BASENAME, (char *) ((LPCTSTR) m_basename)); }
void CHeadtrackerDlg::OnKillfocusFirst() { PrintEditNumber4D(IDC_FIRST, m_first_filenum); }
void CHeadtrackerDlg::OnKillfocusLast() { PrintEditNumber4D(IDC_LAST, m_last_filenum); }

void CHeadtrackerDlg::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pSB) 
{
	int n;

	if (pSB->GetDlgCtrlID() == IDC_SCROLLBAR_SPEED) {

		n = pSB->GetScrollPos();
		switch(nSBCode) {
		case SB_THUMBPOSITION:
			n = nPos;
			break;
		case SB_LINELEFT: // left arrow button
			n = max(n-1, 0);
			break;
		case SB_LINERIGHT: // right arrow button
			n = min(n+1, N_DISPLAY_SPEEDS-1);
			break;
		case SB_ENDSCROLL:
			// For some reason, the function is called twice after the user clicks --
			// the second invocation is with SB_ENDSCROLL and must be ignored.
			return;
			break;
		}
		pSB->SetScrollPos(n);
		display_speed = display_speeds[n];
		PrintDisplaySpeed(display_speed);
		if (KillTimer(TIMER_ID) == 0)
			AfxMessageBox("Failed to Kill Timer",MB_OK|MB_ICONSTOP,0);
		else {
			if (SetTimer(TIMER_ID, 1000/display_speeds[n], NULL) == 0)
				AfxMessageBox("Failed to Set Timer",MB_OK|MB_ICONSTOP,0);
		}
	}
}


/////////////////////////////////////////////////////////////////////////////
// Camera functions

void CHeadtrackerDlg::OnCameraOff() { UpdateData(TRUE);	StopCamera(); }
void CHeadtrackerDlg::OnCameraOn() { UpdateData(TRUE); }
void CHeadtrackerDlg::OnZoom() { NegateBoolean(&zoom_control); }
void CHeadtrackerDlg::OnCameraHome() { SendCameraHome(); }
void CHeadtrackerDlg::OnLockWhiteBalance() { CameraLockWhiteBalance(); }


/////////////////////////////////////////////////////////////////////////////
// Monitoring functions

FILE *fp_manual = NULL;

void CHeadtrackerDlg::OnShowPixelLists() { ShowPixelListsAndOutlines(); }
void CHeadtrackerDlg::OnSaveErrorFunction() { SaveErrorFunctions(); }

void CHeadtrackerDlg::OnLButtonDown(UINT nFlags, CPoint point) 
{
	int skipx = LEFT_MARGIN+ISIZEX, skipy = TOP_MARGIN+ISIZEY;
	int x, y;
	unsigned char *ptr;
	int col1, col2, col3;

	x = point.x - skipx * (point.x / skipx) - LEFT_MARGIN;
	y = point.y - skipy * (point.y / skipy) - TOP_MARGIN;
	ptr = Image.GetImagePtr();
	ptr += 3*(y*ISIZEX+x);
	sprintf(msg, "(%3d,%3d)", x, y);
	GetDlgItem(IDC_XY)->SetWindowText(msg);
	sprintf(msg, "(%3d,%3d,%3d)", *ptr, *(ptr+1), *(ptr+2));
	GetDlgItem(IDC_BGR)->SetWindowText(msg);
	PrintBinContents(*ptr, *(ptr+1), *(ptr+2));
	BGRToColorBins(*ptr, *(ptr+1), *(ptr+2), &col1, &col2, &col3);
	sprintf(msg, "(%3d,%3d,%3d)", col1, col2, col3);
	GetDlgItem(IDC_COLOR_BINS)->SetWindowText(msg);

	if (fp_manual) 
		fprintf(fp_manual, "%3d %3d %3d ; ...\n", m_filenum, x, y);
}

void CHeadtrackerDlg::OnTimeSequence() 
{
	struct _timeb timebuffer1, timebuffer2;
	unsigned char *images;
	int i_rel;
	int niter = 0;
	int i;
	int j;

	if (m_input_mode == IMAGE_SEQUENCE)  {
		img_reddot.Display(my_hdc,TIME_DOT_X, TIME_DOT_Y);

		// Allocate memory and read images
		numfiles = m_last_filenum - m_first_filenum + 1;
		images = (unsigned char *) 
			malloc(numfiles*BYTES_PER_PIXEL_IN_SEQUENCE*ISIZEX*ISIZEY*sizeof(unsigned char));
		if (images == NULL) {
			AfxMessageBox("Unable to allocate memory for image sequence",MB_OK|MB_ICONSTOP,0);
			return;
		}
		for (i=m_first_filenum ; i<=m_last_filenum ; i++) {
			sprintf(fname, m_basename, i);
			if (!Image.Read24BitBMP(fname)) {
				AfxMessageBox("Error reading BMP file",MB_OK|MB_ICONSTOP,0);
				break;
			} else {
				i_rel = i - m_first_filenum;
				memcpy(images+i_rel*BYTES_PER_PIXEL_IN_SEQUENCE*ISIZEX*ISIZEY,
					Image.GetImagePtr(), BYTES_PER_PIXEL_IN_SEQUENCE*ISIZEX*ISIZEY);
			}
		}

		// Track throughout sequence while timing
		_ftime( &timebuffer1 );
		for (j=0; j<3;j++) {
			for (i=m_first_filenum ; i<=m_last_filenum ; i++) {
				i_rel = i - m_first_filenum;
				memcpy(Image.GetImagePtr(), 
					images+i_rel*BYTES_PER_PIXEL_IN_SEQUENCE*ISIZEX*ISIZEY, BYTES_PER_PIXEL_IN_SEQUENCE*ISIZEX*ISIZEY);
//				Image.Display(my_hdc,LEFT_MARGIN,TOP_MARGIN);			
//				sprintf(msg, "%4d", i);
//				GetDlgItem(IDC_FRAME_NUMBER)->SetWindowText(msg);
				if (headtrackerDlg->m_tracking)  {
					TrackAndDisplay();
				}
				niter++;
			}
		}
		_ftime( &timebuffer2 );
		img_blankdot.Display(my_hdc,TIME_DOT_X, TIME_DOT_Y);
		float timediff = (timebuffer2.time - timebuffer1.time) + (timebuffer2.millitm - timebuffer1.millitm)/1000.0f;
		sprintf(msg, "Elapsed time: %5.1f seconds.  Result: %3d ms = %5.1f Hz", timediff, (int) (timediff*1000.0f/niter), niter / timediff);
		AfxMessageBox(msg, MB_OK|MB_ICONINFORMATION,0);
		free(images);
	}
}

void CHeadtrackerDlg::OnOpenManual() 
{
	if (fp_manual) fclose(fp_manual);
	fp_manual = fopen("../../seq_results/locm.m", "w");
	if (fp_manual == NULL) {
		AfxMessageBox("Error opening 'locm.m' file", MB_OK|MB_ICONSTOP,0);
	} else {
		fprintf(fp_manual, "loc = [ ...\n");
	}
}

void CHeadtrackerDlg::OnCloseManual() 
{
	if (fp_manual != NULL) {
		fprintf(fp_manual, "];\n");
		fclose(fp_manual);
		fp_manual = NULL;
	}
}

void CHeadtrackerDlg::OnOverlayPrecomputedEllipses() 
{
	FILE *fp;
	EllipseState state, state2;
	BOOL error = FALSE;
	char fname_out[80];
	int i, frame;

	if (m_input_mode == IMAGE_SEQUENCE)  {
		sprintf(msg, "../../seq1/%s", fname_ellipse_states);
		fp = fopen(msg, "r");
		if (fp == NULL) {
			sprintf(msg, "Error opening '%s' file", fname_ellipse_states);
			AfxMessageBox(msg, MB_OK|MB_ICONSTOP,0);
		} else {
			fgets(msg, 80, fp);	// ignore first line
			img_blankdot.Display(my_hdc,TRACKING_DOT_X, TRACKING_DOT_Y);
			for (i=m_first_filenum ; i<=m_last_filenum && !error ; i++) {
				sprintf(fname, m_basename, i);
				if (!Image.Read24BitBMP(fname)) {
					AfxMessageBox("Error reading BMP file",MB_OK|MB_ICONSTOP,0);
					break;
				} else {
					fscanf(fp, "%d %d %d %d\n", &frame, &(state.x), &(state.y), &(state.sz));
					if (frame != i)  {
						AfxMessageBox("Something wrong with ellipse_states file (1)",MB_OK|MB_ICONSTOP,0);
						error = TRUE;
					}
					Image.Display(my_hdc,LEFT_MARGIN,TOP_MARGIN);			
					sprintf(msg, "%4d", i);
					GetDlgItem(IDC_FRAME_NUMBER)->SetWindowText(msg);
					state2.x = state.x;  state2.y = state.y;  state2.sz = state.sz; 
					KeepStateInBounds(&state);
					if (state2.x != state.x || state2.y != state.y || state2.sz != state.sz) {
						AfxMessageBox("Something wrong with ellipse_states file (2)",MB_OK|MB_ICONSTOP,0);
						error = TRUE;
					}
					Image.ConvertToImage8(&img_gray);
#if 1 // overlay white outline => 8-bit BMP image
					OverlayOutline8(&img_gray, &state, 255, &img_outline8);
					sprintf(fname_out, "../../seq_results/overlay%03d.bmp", i);
					if (!img_outline8.Write8BitBMP(fname_out)) {
#else // overlay red outline => 24-bit BMP image
					OverlayOutlineBGR24(&img_gray, &state, 255, 0, 0, &img_outline);
					sprintf(fname_out, "../../seq_results/overlay%03d.bmp", i);
					if (!img_outline.Write24BitBMP(fname_out)) {
#endif
						AfxMessageBox("Error writing BMP file",MB_OK|MB_ICONSTOP,0);
						break;
					};
				}
			}
			img_reddot.Display(my_hdc,TRACKING_DOT_X, TRACKING_DOT_Y);
			fclose(fp);
		}
	}	
}

