#include "stdafx.h"	// TRACE(), AfxMessageBox(), max(), min()
#include "assert.h"

#include "base.h" // ISIZE[XY]
#define VARSIZE // prevents "filters.h" from defining ISIZEX & ISIZEY
#include "filters.h"
#include "draw.h"
#include "gradient_sum.h"
#include "color_histogram.h"
#include "outline.h"
#include "search.h" // EllipseState
#include "track.h"

// #define COMPUTE_PIXEL_LISTS   // forces computation rather than reading precomputed data

// images
Image8 img_gray(ISIZEX,ISIZEY), img_gray_old(ISIZEX,ISIZEY);
Image16 img_gauss(ISIZEX,ISIZEY);
Image8 img_grad(ISIZEX,ISIZEY);
Image16 img_gradx(ISIZEX,ISIZEY);
Image16 img_grady(ISIZEX,ISIZEY);
Image8 img_gradx8(ISIZEX,ISIZEY);
Image8 img_grady8(ISIZEX,ISIZEY);
ImageFloat img_gradxf(ISIZEX,ISIZEY);
ImageFloat img_gradyf(ISIZEX,ISIZEY);
Image16 img_tmp16(ISIZEX,ISIZEY);
ImageBGR24 img_outline(ISIZEX,ISIZEY);
ImageBGR24 img_outline_grad(ISIZEX,ISIZEY);
ImageBGR24 img_outline_color(ISIZEX,ISIZEY);
Image8 img_outline8(ISIZEX,ISIZEY);
Image8 img_outline_grad8(ISIZEX,ISIZEY);
Image8 img_outline_color8(ISIZEX,ISIZEY);
Image8 img_mask(ISIZEX,ISIZEY);
Image8 img_color1(ISIZEX,ISIZEY);
Image8 img_color2(ISIZEX,ISIZEY);
Image8 img_color3(ISIZEX,ISIZEY);
Image8 img_color1ref(ISIZEX,ISIZEY);
Image8 img_color2ref(ISIZEX,ISIZEY);
Image8 img_color3ref(ISIZEX,ISIZEY);
Image8 img_color1q(ISIZEX,ISIZEY);
Image8 img_color2q(ISIZEX,ISIZEY);
Image8 img_color3q(ISIZEX,ISIZEY);
Image8 img_likelihood(ISIZEX,ISIZEY);
Image8 img_backprojection(ISIZEX,ISIZEY);
Image8 img_lights(ISIZEX,ISIZEY);
Image8 img_bin(ISIZEX,ISIZEY);

// external variables
extern BOOL building_model;
extern HDC my_hdc;  // for debugging, to allow displaying of images
extern CDC *my_cdc;  // for debugging, to allow TextOut()
extern BOOL use_gradient;
extern BOOL use_color;
extern BOOL adapt_histogram;
extern BOOL show_details;
extern BOOL sum_gradient_magnitude;

// Other variables
ColorHistogram ch_model;
ColorHistogram ch_ref;
ColorHistogram ch_background;
ColorHistogram ch_curr;
int n_images_in_model;
int color_bin_of_interest=0;
static char msg[80];
int tmp1, tmp2;
int tmp3;

// Outlines
OUTLINE outlines[MAX_OUTLINE_SZ + 1];

// Search strategy
SearchStrategy search_strategy;
SearchStrategy ss_tmp;

// Search range is +/- {msx, msy} pixels
int msx = 16, msy = 4;

// Skip szinc pixels to find the smaller and larger scale
int szinc = 1;


// just for debugging and display
void Add32768(Image16 *img)
{
	short *ptri = (short*) img->GetImagePtr();
	unsigned short *ptro = img->GetImagePtr();
	int val;
	int i;

	for (i=0 ; i<ISIZEX*ISIZEY ; i++) {
		val = *ptri++;
		*ptro++ = val + 32768;
	}
}

void MakeSearchStrategy(SearchStrategy *ss)
{
	int i = 0;
	int x, y, sz;
	BOOL left_to_right = TRUE;
	BOOL top_to_bottom = TRUE;

	assert(N_STATE_VARIABLES == 3);
	for (sz = -szinc ; sz <= szinc ; sz += szinc) {
		for (y = -msy ; y <= msy ; y++) {
			for (x = -msx ; x <= msx ; x++) {
				assert(i < N_STATE_VARIABLES * MAX_N_STATES_TO_CHECK - 2);
				assert(x>=-127 && x<=127);
				assert(y>=-127 && y<=127);
				assert(sz>=-127 && sz<=127);
				ss->ds[i++] = (searchLocType) (left_to_right ? x : -x);
				ss->ds[i++] = (searchLocType) (top_to_bottom ? y : -y);
				ss->ds[i++] = (searchLocType) sz;
//				TRACE("     check: dx dy ds %3d %3d %3d\n",
//					ss->ds[i-3], ss->ds[i-2], ss->ds[i-1]);
			}
			left_to_right = 1 - left_to_right;
		}
		top_to_bottom = 1 - top_to_bottom;
	}

	assert(i%N_STATE_VARIABLES==0);
	ss->n_states_to_check = i/N_STATE_VARIABLES;
}

// AndNotBitmap()
// Returns img1 && !(img2)
static void AndNotBitmap(Image8* img1, Image8* img2, Image8* output)
{
	unsigned char *ptr1, *ptr2, *ptrout;
	int i;

	ptr1 = img1->GetImagePtr();
	ptr2 = img2->GetImagePtr();
	ptrout = output->GetImagePtr();

	for (i = 0 ; i < ISIZEX*ISIZEY ; i++) {
		*ptrout++ = 255 * (*ptr1 && !(*ptr2));
		ptr1++; ptr2++;
	}
}

void ComputePixelList(Image8 *img1,
					  Image8 *img2,
					  int xcen, int ycen,
					  PixelList *pl
					  )
{
	Image8 img_diff(ISIZEX,ISIZEY);
	unsigned char *ptr;
	pixelListType *ptro;
	int x, y;
	int i=0;

	AndNotBitmap(img1, img2, &img_diff);

	// Count no. of nonzero pixels
	ptr = img_diff.GetImagePtr();
	for (y=0 ; y<ISIZEX*ISIZEY ; y++) {
		if (*ptr++)  i++;
	}
	pl->n_pixels = i;
	pl->pixels = (pixelListType *) malloc(pl->n_pixels * 2 * sizeof(pixelListType));
	
	// Copy nonzero pixels to list
	ptr = img_diff.GetImagePtr();
	ptro = pl->pixels;
	for (y=0 ; y<ISIZEY ; y++) {
		for (x=0 ; x<ISIZEX ; x++) {
			if (*ptr++) {
				assert(x-xcen>=-128 && x-xcen<=127);
				assert(y-ycen>=-128 && y-ycen<=127);
				*ptro++ = (pixelListType) (x - xcen);
				*ptro++ = (pixelListType) (y - ycen);
			}
		}
	}
	
}

void** createArray2D(int ncols, int nrows, int nbytes)
{
	char **tt;
	int i;

     tt = (char **) malloc(nrows * sizeof(void *) +
                           ncols * nrows * nbytes);
	 if (tt == NULL) AfxMessageBox("(createArray2D) out of memory!!", MB_OK|MB_ICONSTOP, 0);	

     for (i = 0 ; i < nrows ; i++)
          tt[i] = ((char *) tt) + (nrows * sizeof(void *) + 
							i * ncols * nbytes);

	return((void **) tt);
}

void PixelList2DArrayAlloc(PixelList ***ptr, int n_states_to_check)
{
	*ptr = (PixelList **) createArray2D(n_states_to_check, MAX_OUTLINE_SZ+1, sizeof(PixelList));
}

/////////////////////////////////////////////////////////////////////////////
// ComputeDifferentialPixelLists()
// Given an ellipse size and a search strategy, fills two arrays
// of pixel lists.
//   add_list contains all the pixels in img2 but not in img1, while
//   sub_list contains all the pixels in img1 but not in img2.
void ComputeDifferentialPixelLists(int sz, 
								   SearchStrategy *ss,
								   PixelList add_list[],
								   PixelList sub_list[]
								   ) 
{
	Image8 img1(ISIZEX,ISIZEY), img2(ISIZEX,ISIZEY);
	searchLocType *ptr = ss->ds;
	int xcen=ISIZEX/2, ycen=ISIZEY/2;
	int x1=xcen, y1=ycen, sz1=sz, x2=xcen, y2=ycen, sz2=sz;
	int i;

	assert(N_STATE_VARIABLES == 3);
	for (i=0 ; i<ss->n_states_to_check ; i++) {
		x2  = xcen + *ptr++;
		y2  = ycen + *ptr++;
		sz2 = sz   + *ptr++;

		// Do not perform computation when sz is out of bounds; this 
		// incompleteness is okay because it will never be needed at
		// runtime.
		if (sz1>=MIN_OUTLINE_SZ && sz1<=MAX_OUTLINE_SZ &&
			sz2>=MIN_OUTLINE_SZ && sz2<=MAX_OUTLINE_SZ) {

			// Center ellipses onto image, then draw
			DrawMask(x1, y1, &outlines[sz1], &img1);
			DrawMask(x2, y2, &outlines[sz2], &img2);

			// Compute two pixel lists
			ComputePixelList(&img2, &img1, xcen, ycen, &add_list[i]);
			ComputePixelList(&img1, &img2, xcen, ycen, &sub_list[i]);
		} else {
			add_list[i].n_pixels = -1;
			sub_list[i].n_pixels = -1;
		}

		x1=x2;  y1=y2;  sz1=sz2;
	}

}

#define PIXELLIST_HEADER_LENGTH 3
char pixellist_header[PIXELLIST_HEADER_LENGTH+1] = "PL1";

void WriteDifferentialPixelLists(char *fname, PixelList **lists,
								 SearchStrategy *ss) 
{
	searchLocType *ptr = ss->ds;
	int i, sz;
	int n_pix;
	int tmp;
	FILE *fp;

	// Open file for writing
	fp = fopen(fname, "wb");
	if (fp == NULL)
		AfxMessageBox("Error: could not open pixel list file for writing!!", MB_OK|MB_ICONSTOP, 0);

	// Write header
	fwrite(pixellist_header, sizeof(char), PIXELLIST_HEADER_LENGTH, fp);

	// Write search strategy
	tmp = N_STATE_VARIABLES;
	fwrite(&tmp, sizeof(int), 1, fp);
	fwrite(&(ss->n_states_to_check), sizeof(int), 1, fp);
	for (i=0 ; i<ss->n_states_to_check ; i++) {
		fwrite(ptr, sizeof(searchLocType), 1, fp);  ptr++;
		fwrite(ptr, sizeof(searchLocType), 1, fp);  ptr++;
		fwrite(ptr, sizeof(searchLocType), 1, fp);  ptr++;
	}

	// Write pixel list
	for (sz=MIN_OUTLINE_SZ ; sz<=MAX_OUTLINE_SZ ; sz++) {
		for (i=0 ; i<ss->n_states_to_check ; i++) {
			n_pix = lists[sz][i].n_pixels;
			fwrite(&n_pix, sizeof(int), 1, fp);
			if (n_pix>=0) {
				tmp = fwrite(lists[sz][i].pixels, sizeof(pixelListType), 2*n_pix, fp);	
				if (tmp != 2*n_pix) 
					AfxMessageBox("Error writing pixel list!!", MB_OK|MB_ICONSTOP, 0);
			}
		}
	}
	fclose(fp);
}

// Reads data and allocates memory
void ReadDifferentialPixelLists(char *fname, 
								PixelList ***lists,
								SearchStrategy *ss) 
{
	char strtmp[PIXELLIST_HEADER_LENGTH+1];
	searchLocType *ptr = ss->ds;
	int i, sz;
	int n_pix;
	int tmp;
	FILE *fp;

	if (*lists != NULL)
		AfxMessageBox("Warning: trying to read pixel list into non-empty array!!", MB_OK|MB_ICONEXCLAMATION, 0);

	// Open file for reading
	fp = fopen(fname, "rb");
	if (fp == NULL)
		AfxMessageBox("Error: could not open pixel list file for reading!!", MB_OK|MB_ICONSTOP, 0);

	// Read header
	fread(strtmp, sizeof(char), PIXELLIST_HEADER_LENGTH, fp);
	strtmp[PIXELLIST_HEADER_LENGTH] = '\0';
	if (strcmp(strtmp, pixellist_header) != 0)
		AfxMessageBox("Error: pixel list file has invalid header!!", MB_OK|MB_ICONSTOP, 0);

	// Read search strategy
	fread(&tmp, sizeof(int), 1, fp);
	if (tmp != N_STATE_VARIABLES)
		AfxMessageBox("Error: pixel list file contains wrong # of state variables!!", MB_OK|MB_ICONSTOP, 0);
	fread(&(ss->n_states_to_check), sizeof(int), 1, fp);
	if (ss->n_states_to_check != (2*msx+1)*(2*msy+1)*3)
		AfxMessageBox("Warning: pixel list file contains questionable # of states to check!!", MB_OK|MB_ICONEXCLAMATION, 0);
	for (i=0 ; i<ss->n_states_to_check ; i++) {
		fread(ptr, sizeof(searchLocType), 1, fp);  ptr++;
		fread(ptr, sizeof(searchLocType), 1, fp);  ptr++;
		fread(ptr, sizeof(searchLocType), 1, fp);  ptr++;
	}

	// Allocate memory for pixel list array
	PixelList2DArrayAlloc(lists, ss->n_states_to_check);

	// Read pixel list
	for (sz=MIN_OUTLINE_SZ ; sz<=MAX_OUTLINE_SZ ; sz++) {
		for (i=0 ; i<ss->n_states_to_check ; i++) {
			fread(&n_pix, sizeof(int), 1, fp);
			(*lists)[sz][i].n_pixels = n_pix;
			if (n_pix>=0) {
				(*lists)[sz][i].pixels = (pixelListType *) malloc(2*n_pix*sizeof(pixelListType));
				tmp = fread((*lists)[sz][i].pixels, sizeof(pixelListType), 2*n_pix, fp);	
				if (feof(fp))
					AfxMessageBox("END OF FILE ENCOUNTERED!!", MB_OK|MB_ICONSTOP, 0);
				if (tmp != 2*n_pix) 
					AfxMessageBox("Error reading pixel list!!", MB_OK|MB_ICONSTOP, 0);
			}
		}
	}
	fclose(fp);
}
#undef PIXELLIST_HEADER_LENGTH

void CompareDifferentialPixelLists(PixelList **lists1,
								   PixelList **lists2,
								   SearchStrategy *ss1,
								   SearchStrategy *ss2) 
{
	int i, j, sz;
	int n_pix1, n_pix2;
	pixelListType *ptr1, *ptr2;

	// Compare search strategies
	if (ss1->n_states_to_check != ss2->n_states_to_check)
		AfxMessageBox("Error: 'n_states_to_check's are different!!", MB_OK|MB_ICONSTOP, 0);
	for (i=0 ; i<N_STATE_VARIABLES * ss1->n_states_to_check ; i++)
		if (ss1->ds[i] != ss2->ds[i])
			AfxMessageBox("Error: Search strategies are different!!", MB_OK|MB_ICONSTOP, 0);

	// Compare pixel lists
	for (sz=MIN_OUTLINE_SZ ; sz<MAX_OUTLINE_SZ ; sz++) {
		for (i=0 ; i<ss1->n_states_to_check ; i++) {
			n_pix1 = lists1[sz][i].n_pixels;
			n_pix2 = lists2[sz][i].n_pixels;
			if (n_pix1 != n_pix2) 
				AfxMessageBox("Error: PixelLists are different!!", MB_OK|MB_ICONSTOP, 0);
			ptr1 = lists1[sz][i].pixels;
			ptr2 = lists2[sz][i].pixels;
			for (j=0 ; j<2*n_pix1 ; j++) {
				if (*ptr1++ != *ptr2++)
				AfxMessageBox("Error: PixelLists are different!!", MB_OK|MB_ICONSTOP, 0);
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////
// smoothImage
// 
// Smooths an image by convolving with a 5x5 Gaussian, using
// two fixed one-dimensional kernels:  [1 4 6 4 1].  After
// the full convolution, the result is divided by 8, so that the
// maximum pixel value is 8192.
// 
// INPUTS
// imagebuf:  grey-scale image
// ncols, nrows:  size of image
// 
// OUTPUTS
// gaussbuf:  smoothed image
// 
// NOTES
// Sets the two-pixel out-of-bounds border to zero.
//
// THIS FUNCTION WAS NOT PART OF THE ORIGINAL CODE.  IT HAS BEEN
// COPIED OVER FROM A RELATED PIECE OF CODE SIMPLY TO MAKE IT EASIER
// FOR THE USER TO REPLACE THE PROPRIETARY LIBRARY CALL 
// SmoothByConvolvingWithGaussian().

void smoothImage(unsigned char *imagebuf, unsigned short *gaussbuf,
			  int ncols, int nrows)
{
	unsigned char *cptr;
	unsigned short *sptr;
	int i, j, indx;

	// Zero top border
	sptr = gaussbuf;
	for (i = 0 ; i < 2*ncols + 2 ; i++)  
		*sptr++ = 0;

	// Convolve middle with horizontal [1 4 6 4 1] mask
	cptr = imagebuf;
	sptr = gaussbuf + 2*ncols + 2;
	for (i = 2 ; i < (nrows-2) ; i++)  {
		for (j = 2 ; j < ncols - 2 ; j++)  {
			indx = i*ncols+j;
			*sptr++ = cptr[indx-2] + 4 * cptr[indx-1] + 6 * cptr[indx]
					+ 4 * cptr[indx+1] + cptr[indx+2];
 		}
		for (j = 1 ; j < 5 ; j++)
			*sptr++ = 0;
	}

	// Convolve middle with vertical [1 4 6 4 1]^T mask
	cptr = imagebuf;
	sptr = gaussbuf + 2*ncols + 2;
	for (i = 2 ; i < (nrows-2) ; i++)  {
		for (j = 2 ; j < ncols - 2 ; j++)  {
			indx = i*ncols+j;
			*sptr += cptr[indx-2*ncols] + 4 * cptr[indx-1*ncols] 
					+ 6 * cptr[indx]
					+ 4 * cptr[indx+1*ncols] + cptr[indx+2*ncols];
			*sptr++ /= 8;
 		}
		sptr += 4;
	}

	// Zero bottom border
	cptr = imagebuf + (nrows-2)*ncols + 2;
	sptr = gaussbuf + (nrows-2)*ncols + 2;
	for (i = (nrows-2)*ncols + 2 ; i < ncols*nrows ; i++)
		*sptr++ = 0;
}


//////////////////////////////////////////////////////////////////////
// computeGradient
// 
// Takes the absolute value of the derivative of an image by
// convolving the image with two one-dimensional filters ([-1 0 1]
// and [-1 0 1]^T), taking the absolute values, and summing.
//  
// INPUTS
// gauss:  smoothed image, as returned by smoothImage()
// ncols, nrows:  size of image
// 
// OUTPUTS
// grad:  absolute derivative
// 
// NOTES
// Sets the two-pixel out-of-bounds border to zero.
// Saturates the values above thresh_absderiv; that is, any value
// above thresh_absderiv becomes 255.
//
// THIS FUNCTION WAS NOT PART OF THE ORIGINAL CODE.  IT HAS BEEN
// COPIED OVER FROM A RELATED PIECE OF CODE SIMPLY TO MAKE IT EASIER
// FOR THE USER TO REPLACE THE PROPRIETARY LIBRARY CALLS 
// GradientMagnitude(), GradientMagnitudeX(), and GradientMagnitudeY().

void computeGradient(unsigned short *gauss, unsigned char *grad,
				 int ncols, int nrows)
{
	register unsigned short *ptrL, *ptrR, *ptrU, *ptrD;
	register unsigned char *ptro;
	register unsigned char *end;
	register int val, row;


	// Zero top border
	for (ptro = grad ; ptro < grad+3*ncols+3 ; *ptro++ = 0) ;

	end = grad + ncols*(nrows-3) - 3;
	for (ptrL = gauss+3*ncols+2, ptrR = gauss+3*ncols+4,
		 ptrU = gauss+2*ncols+3, ptrD = gauss+4*ncols+3 ;
		 ptro < end; )  {

		val = abs(*ptrL++ - *ptrR++) + abs(*ptrU++ - *ptrD++);
		if (val > thresh_absderiv)  val = 255;

		// Clear left and right three-pixel borders
		row = (ptro - grad) % ncols;
		if (row <= 2 || row >= ncols - 3)  val = 0;

		assert(ptro >= grad);
		assert(ptro < grad + ncols*nrows*sizeof(char));
		*ptro++ = val;
	}

	// Zero bottom border
	for ( ; ptro < grad+ncols*nrows ; *ptro++ = 0) ;
}

//////////////////////////////////////////////////////////////////////
// searchForHead
//
// INPUTS
// s_old:  old state (x,y,sz)
//
// OUTPUTS
// s_new:  new state (x,y,sz)
int grad_scores[MAX_N_STATES_TO_CHECK];
int color_scores[MAX_N_STATES_TO_CHECK];

void SearchForHead(EllipseState *s_old, ImageBGR24 *img, 
				   SearchStrategy *ss, 
				   PixelList addlist[],
				   PixelList sublist[],
				   EllipseState *sgrad,
				   EllipseState *scolor,
				   EllipseState *s_new)
{
	int x;
	int y;
	int sz;
	int sz_curr;
	searchLocType *ptr = ss->ds;
	int best_score_grad = -1, best_score_color = -1, best_score = -1;
	int worst_score_grad = 999999, worst_score_color = 999999;
	ColorHistogram ch_ref_scaled, ch_new, ch_tmp;
	int i;
//	int tmp;
	int curr_score;
//	int n_saturated_pixels;

	assert(s_old->sz>MIN_OUTLINE_SZ && s_old->sz<MAX_OUTLINE_SZ);
	assert(s_old->x >= msx + outlines[s_old->sz+1].halfsize_x);
	assert(s_old->x <= ISIZEX-1 - msx - outlines[s_old->sz+1].halfsize_x);
	assert(s_old->y >= msy + outlines[s_old->sz+1].halfsize_y);
	assert(s_old->y <= ISIZEY-1 - msy - outlines[s_old->sz+1].halfsize_y);
	assert(ss->n_states_to_check <= MAX_N_STATES_TO_CHECK);

	// Preprocessing
	img->ConvertToImage8(&img_gray);
	if (use_gradient) {
		SmoothByConvolvingWithGaussian(&img_gray, &img_gauss, 5.0f);
		if (sum_gradient_magnitude) {
			GradientMagnitude(&img_gauss, &img_tmp16);
			img_tmp16.ConvertToImage8(8, &img_grad);
		} else {
			GradientMagnitudeX(&img_gauss, &img_gradx);
			GradientMagnitudeY(&img_gauss, &img_grady);
			img_gradx.ConvertToImageFloat(FALSE, &img_gradxf); // FALSE indicates that img_gradx is 'short', as opposed to 'unsigned short'
			img_grady.ConvertToImageFloat(FALSE, &img_gradyf);
			if (show_details) {  // This is simply for debugging purposes
				Add32768(&img_gradx);
				Add32768(&img_grady);
				img_gradx.ConvertToImage8(8, &img_gradx8);
				img_grady.ConvertToImage8(8, &img_grady8);
			}
		}
	}
	if (use_color) {
		ExtractColorSpace(img, &img_color1, &img_color2, &img_color3);
		// QuantizeColorSpace(&img_color1, &img_color2, &img_color3, &img_color1q, &img_color2q, &img_color3q);
	}

	// Evaluate states based on normalized gradient
	if (use_gradient) {
		if (sum_gradient_magnitude) {
			for (i = 0 ; i < ss->n_states_to_check ; i++) {
				x  = s_old->x  + *ptr++;  y  = s_old->y  + *ptr++;  sz = s_old->sz + *ptr++;
				grad_scores[i]  = NormalizedSumOfGradientMagnitude(x, y, &outlines[sz], img_grad.GetImagePtr());
			}
		} else { // sum dot product
			for (i = 0 ; i < ss->n_states_to_check ; i++) {
				x  = s_old->x  + *ptr++;  y  = s_old->y  + *ptr++;  sz = s_old->sz + *ptr++;
				grad_scores[i]  = NormalizedSumOfGradientDotProduct(x, y, &outlines[sz], 
						img_gradxf.GetImagePtr(), img_gradyf.GetImagePtr());
			}
		}
	}

	// Display likelihood
	if (use_color && show_details) {
		DrawColorLikelihoodMap(img, &ch_ref, &img_likelihood);
		ComputeColorHistogramOfWholeImage(&img_color1, &img_color2, &img_color3, &ch_background);
		DrawHistogramBackprojection(img, &ch_ref, &ch_background, &img_backprojection);
	}
	
	// Evaluate states based on color histogram
	if (use_color) {
		x = s_old->x;  y = s_old->y;  sz = s_old->sz;
		sz_curr = sz+ss->ds[2];
		ComputeColorHistogram(&img_color1, &img_color2, &img_color3, x+ss->ds[0], y+ss->ds[1], 
							&outlines[sz_curr], &ch_new);
		assert(SizeOfColorHistogram(&ch_new) == outlines[sz_curr].area);
		NormalizeColorHistogram(&ch_ref, outlines[sz_curr].area, 
								outlines[CH_MODEL_SZ].area, &ch_ref_scaled);
		curr_score = ColorHistogramIntersection(&ch_ref_scaled, &ch_new);
		color_scores[0] = (curr_score * 10000) / outlines[sz_curr].area;

		for (i = 1 ; i < ss->n_states_to_check ; i++) {
			if (sz + ss->ds[N_STATE_VARIABLES*i+2] != sz_curr) {
				sz_curr = sz + ss->ds[N_STATE_VARIABLES*i+2];
				NormalizeColorHistogram(&ch_ref, outlines[sz_curr].area, 
								outlines[CH_MODEL_SZ].area, &ch_ref_scaled);
				UpdateColorHistogram(&img_color1, &img_color2, &img_color3, x, y, &addlist[i], &sublist[i], &ch_new);
				assert(SizeOfColorHistogram(&ch_new) == outlines[sz_curr].area);
				curr_score = ColorHistogramIntersection(&ch_ref_scaled, &ch_new);
				color_scores[i] = (curr_score * 10000) / outlines[sz_curr].area;
			} else {
				UpdateColorHistogramAndIntersection(&img_color1, &img_color2, &img_color3, x, y, 
								&addlist[i], &sublist[i], &ch_ref_scaled, &ch_new, &curr_score);
				assert(SizeOfColorHistogram(&ch_new) == outlines[sz_curr].area);
				color_scores[i] = (curr_score * 10000) / outlines[sz_curr].area;
			}

#if 0 // useful for debugging
			ComputeColorHistogram(&img_color1, &img_color2, &img_color3, x+ss->ds[N_STATE_VARIABLES*i], y+ss->ds[N_STATE_VARIABLES*i+1], 
							&outlines[sz_curr], &ch_tmp);
			tmp = AreColorHistogramsEqual(&ch_new, &ch_tmp);
			if (!tmp) {
				my_cdc->TextOut(200, 215, "Color Histograms are Not Equal!!");
			}
			tmp = ColorHistogramIntersection(&ch_ref_scaled, &ch_tmp);
			if (tmp != curr_score) {
				my_cdc->TextOut(200, 215, "Current score is incorrect!!");
			}
#endif
		}
	}

	// Compute best and worst scores for grad and color alone
	for (i = 0 ; i < ss->n_states_to_check ; i++) {
		if (use_gradient) {
			curr_score = grad_scores[i];
			if (curr_score > best_score_grad) {
				best_score_grad = curr_score;
				sgrad->x  = s_old->x  + ss->ds[N_STATE_VARIABLES*i];
				sgrad->y  = s_old->y  + ss->ds[N_STATE_VARIABLES*i+1];
				sgrad->sz = s_old->sz + ss->ds[N_STATE_VARIABLES*i+2];
			}
			if (curr_score < worst_score_grad) {
				worst_score_grad = curr_score;
			}
		}
		if (use_color) {
			curr_score = color_scores[i];
			if (curr_score > best_score_color) {
				best_score_color = curr_score;
				scolor->x  = s_old->x  + ss->ds[N_STATE_VARIABLES*i];
				scolor->y  = s_old->y  + ss->ds[N_STATE_VARIABLES*i+1];
				scolor->sz = s_old->sz + ss->ds[N_STATE_VARIABLES*i+2];
			}
			if (curr_score < worst_score_color) {
				worst_score_color = curr_score;
			}
		}
	}

	// Select best state
	float grad_fact  = 1000.0f/(best_score_grad  - worst_score_grad);
	float color_fact = 1000.0f/(best_score_color - worst_score_color);
	if (!use_gradient) grad_fact = 0.0f;
	if (!use_color) color_fact = 0.0f;
//	if (n_saturated_pixels >= 500)  {tmp1=n_saturated_pixels; color_fact *= 0.2f;} else {tmp1=0;}
	for (i = 0 ; i < ss->n_states_to_check ; i++) {
		curr_score = (building_model ? grad_scores[i] : 
			(int) ((grad_scores[i]  - worst_score_grad)  * grad_fact +
			       (color_scores[i] - worst_score_color) * color_fact));
		if (curr_score > best_score) {
			best_score = curr_score;
			s_new->x  = s_old->x  + ss->ds[N_STATE_VARIABLES*i];
			s_new->y  = s_old->y  + ss->ds[N_STATE_VARIABLES*i+1];
			s_new->sz = s_old->sz + ss->ds[N_STATE_VARIABLES*i+2];
		}
	}

//	sprintf(msg, "using grad %d %s, using color %d    ",
//		use_gradient, sum_gradient_magnitude?"MAG":"DOT", use_color);
//	my_cdc->TextOut(5, 225, msg);

//	sprintf(msg, "grad: <%4d,%4d>, color: <%4d,%4d>       ", 
//		worst_score_grad, best_score_grad, worst_score_color, best_score_color);
//	my_cdc->TextOut(5, 225, msg);

	// Build model, if desired
	// Model is always of size CH_MODEL_SZ
	if (building_model || adapt_histogram) {
		ComputeColorHistogram(&img_color1, &img_color2, &img_color3, s_new->x, s_new->y,
							&outlines[s_new->sz], &ch_new);
		assert(SizeOfColorHistogram(&ch_new) == outlines[s_new->sz].area);
		NormalizeColorHistogram(&ch_new, outlines[CH_MODEL_SZ].area,
								outlines[s_new->sz].area, &ch_tmp);
	}
	if (building_model) {
		AddColorHistograms(&ch_model, &ch_tmp);
		memcpy(ch_ref.val, ch_model.val, N_BINS_TOT*sizeof(int));
		n_images_in_model++;
	}
	if (adapt_histogram) {
		memset(ch_ref.val, 0, N_BINS_TOT*sizeof(int));
		AddColorHistograms(&ch_ref, &ch_model);
		AddColorHistograms(&ch_ref, &ch_tmp);		
		NormalizeColorHistogram(&ch_ref, 1, 2, &ch_ref);
	}

	// When user clicks the mouse, statistics on current histogram can be displayed.
	if (show_details) {
		ComputeColorHistogram(&img_color1, &img_color2, &img_color3, s_new->x, s_new->y,
							&outlines[s_new->sz], &ch_curr);
		DrawPixelsInCertainColorBin(img, &ch_curr, color_bin_of_interest, &img_bin);
	}
}

void KeepStateInBounds(EllipseState *s)
{
	s->sz = max(s->sz, MIN_OUTLINE_SZ+1);
	s->sz = min(s->sz, MAX_OUTLINE_SZ-1);
	s->x  = max(s->x, msx + outlines[s->sz+1].halfsize_x);
	s->x  = min(s->x, ISIZEX-1 - msx - outlines[s->sz+1].halfsize_x);
	s->y  = max(s->y, msy + outlines[s->sz+1].halfsize_y);
	s->y  = min(s->y, ISIZEY-1 - msy - outlines[s->sz+1].halfsize_y);
}

PixelList **add_lists = NULL;
PixelList **sub_lists = NULL;
PixelList **add_lists_tmp = NULL;
PixelList **sub_lists_tmp = NULL;

void InitTracker(EllipseState *curr_state)
{
	static BOOL first_time = TRUE;
	int i;

	if (!first_time) 
		AfxMessageBox("Warning:  InitTracker called more than once!!", MB_OK|MB_ICONEXCLAMATION, 0);	
	first_time = FALSE;

	// Initialize elliptical outlines
	for (i = MIN_OUTLINE_SZ ; i <= MAX_OUTLINE_SZ ; i++)  {
		MakeEllipticalOutline(&outlines[i], i, (int) (1.2 * i));
	}

	/*
	DrawMask(64, 48, &outlines[15], &img_mask); img_mask.Write8BitBMP("img15.bmp");
	DrawMask(64, 48, &outlines[16], &img_mask); img_mask.Write8BitBMP("img16.bmp");
	DrawMask(64, 48, &outlines[17], &img_mask); img_mask.Write8BitBMP("img17.bmp");
	DrawMask(64, 48, &outlines[18], &img_mask); img_mask.Write8BitBMP("img18.bmp");
	DrawMask(64, 48, &outlines[19], &img_mask); img_mask.Write8BitBMP("img19.bmp");
	DrawMask(64, 48, &outlines[20], &img_mask); img_mask.Write8BitBMP("img20.bmp");
	DrawMask(64, 48, &outlines[21], &img_mask); img_mask.Write8BitBMP("img21.bmp");
	DrawMask(64, 48, &outlines[22], &img_mask); img_mask.Write8BitBMP("img22.bmp");
	DrawMask(64, 48, &outlines[23], &img_mask); img_mask.Write8BitBMP("img23.bmp");
	DrawMask(64, 48, &outlines[24], &img_mask); img_mask.Write8BitBMP("img24.bmp");
	DrawMask(64, 48, &outlines[25], &img_mask); img_mask.Write8BitBMP("img25.bmp");
	img_lights = 0;
	curr_state->x = 64;	curr_state->y = 48;
	curr_state->sz = 15;	OverlayOutline8(&img_lights, curr_state, 255, &img_mask);  img_mask.Write8BitBMP("outline15.bmp");
	curr_state->sz = 16;	OverlayOutline8(&img_lights, curr_state, 255, &img_mask);  img_mask.Write8BitBMP("outline16.bmp");
	curr_state->sz = 17;	OverlayOutline8(&img_lights, curr_state, 255, &img_mask);  img_mask.Write8BitBMP("outline17.bmp");
	curr_state->sz = 18;	OverlayOutline8(&img_lights, curr_state, 255, &img_mask);  img_mask.Write8BitBMP("outline18.bmp");
	curr_state->sz = 19;	OverlayOutline8(&img_lights, curr_state, 255, &img_mask);  img_mask.Write8BitBMP("outline19.bmp");
	curr_state->sz = 20;	OverlayOutline8(&img_lights, curr_state, 255, &img_mask);  img_mask.Write8BitBMP("outline20.bmp");
	curr_state->sz = 21;	OverlayOutline8(&img_lights, curr_state, 255, &img_mask);  img_mask.Write8BitBMP("outline21.bmp");
	curr_state->sz = 22;	OverlayOutline8(&img_lights, curr_state, 255, &img_mask);  img_mask.Write8BitBMP("outline22.bmp");
	curr_state->sz = 23;	OverlayOutline8(&img_lights, curr_state, 255, &img_mask);  img_mask.Write8BitBMP("outline23.bmp");
	curr_state->sz = 24;	OverlayOutline8(&img_lights, curr_state, 255, &img_mask);  img_mask.Write8BitBMP("outline24.bmp");
	curr_state->sz = 25;	OverlayOutline8(&img_lights, curr_state, 255, &img_mask);  img_mask.Write8BitBMP("outline25.bmp");
	*/

#ifdef COMPUTE_PIXEL_LISTS
	MakeSearchStrategy(&search_strategy);
	PixelList2DArrayAlloc(&add_lists, search_strategy.n_states_to_check);
	PixelList2DArrayAlloc(&sub_lists, search_strategy.n_states_to_check);
	for (i = MIN_OUTLINE_SZ ; i <= MAX_OUTLINE_SZ ; i++)  {
		ComputeDifferentialPixelLists(i, &search_strategy, add_lists[i], sub_lists[i]); 
	}
	WriteDifferentialPixelLists("lists_add.dat", add_lists, &search_strategy);
	WriteDifferentialPixelLists("lists_sub.dat", sub_lists, &search_strategy);
#else
	if (msx==8 && msy==8) {
		ReadDifferentialPixelLists("lists_add8x8.dat", &add_lists, &search_strategy);
		ReadDifferentialPixelLists("lists_sub8x8.dat", &sub_lists, &search_strategy);
	} else if (msx==4 && msy==4) {
		ReadDifferentialPixelLists("lists_add4x4.dat", &add_lists, &search_strategy);
		ReadDifferentialPixelLists("lists_sub4x4.dat", &sub_lists, &search_strategy);
	} else if (msx==16 && msy==16) {
		ReadDifferentialPixelLists("lists_add16x16.dat", &add_lists, &search_strategy);
		ReadDifferentialPixelLists("lists_sub16x16.dat", &sub_lists, &search_strategy);
	} else if (msx==16 && msy==4) {
		ReadDifferentialPixelLists("lists_add16x4.dat", &add_lists, &search_strategy);
		ReadDifferentialPixelLists("lists_sub16x4.dat", &sub_lists, &search_strategy);
	} else if (msx==2 && msy==2) {
		ReadDifferentialPixelLists("lists_add2x2.dat", &add_lists, &search_strategy);
		ReadDifferentialPixelLists("lists_sub2x2.dat", &sub_lists, &search_strategy);
	} else {
		AfxMessageBox("msx and msy are invalid!!", MB_OK|MB_ICONSTOP, 0);	
		// Read anyway, just so we don't get any memory faults
		ReadDifferentialPixelLists("lists_add8.dat", &add_lists, &search_strategy);
		ReadDifferentialPixelLists("lists_sub8.dat", &sub_lists, &search_strategy);
	}

#endif
//	ReadDifferentialPixelLists("lists_add4x4.dat", add_lists_tmp, &ss_tmp);
//	ReadDifferentialPixelLists("lists_sub4x4.dat", sub_lists_tmp, &ss_tmp);
//	CompareDifferentialPixelLists(add_lists, add_lists_tmp, &search_strategy, &ss_tmp);
//	CompareDifferentialPixelLists(sub_lists, sub_lists_tmp, &search_strategy, &ss_tmp);
//	pixelListType tmpadd[31], tmpsub[37];
//	memcpy(tmpadd, add_lists[24][0].pixels, 31*sizeof(pixelListType));
//	memcpy(tmpsub, sub_lists[24][0].pixels, 37*sizeof(pixelListType));

	curr_state->x = ISIZEX/2;
	curr_state->y = ISIZEY/2;
	curr_state->sz = MAX_OUTLINE_SZ-1;
}

void track(ImageBGR24* img, EllipseState *curr_state)
{
	EllipseState grad_state, color_state, new_state;
	static int xvel = 0, yvel = 0;	// velocities (for prediction)

	curr_state->x += xvel;
	curr_state->y += yvel;
	KeepStateInBounds(curr_state);
	SearchForHead(curr_state, 
				  img, 
				  &search_strategy, 
				  add_lists[curr_state->sz],
				  sub_lists[curr_state->sz],
				  &grad_state,
				  &color_state,
				  &new_state);
//	DrawMask(new_state.x, new_state.y, &outlines[new_state.sz], &img_mask);
	xvel = new_state.x - curr_state->x;
	yvel = new_state.y - curr_state->y;
	curr_state->x = new_state.x;
	curr_state->y = new_state.y;
	curr_state->sz = new_state.sz;
	OverlayOutlineBGR24(&img_gray, curr_state, 255, 0, 0, &img_outline);
	if (show_details) {
		OverlayOutline8(&img_gray, curr_state, 255, &img_outline8);
		if (use_gradient) {
			OverlayOutlineBGR24(&img_gray, &grad_state, 0, 255, 0, &img_outline_grad);
			OverlayOutline8(&img_gray, &grad_state, 255, &img_outline_grad8);
		}
		if (use_color) { 
			OverlayOutlineBGR24(&img_gray, &color_state, 0, 0, 255, &img_outline_color);
			OverlayOutline8(&img_gray, &color_state, 255, &img_outline_color8);
		}
	}
}

void InitBuildingOfModel()
{
//	AfxMessageBox("Model being initialized", MB_OK|MB_ICONSTOP, 0);	
	n_images_in_model = 0;
	memset(ch_model.val, 0, N_BINS_TOT * sizeof(int));
}

void CopyModelHistogramToRef()
{
	memcpy(ch_ref.val, ch_model.val, N_BINS_TOT*sizeof(int));
}

void StopBuildingOfModel()
{
	if (n_images_in_model>0)
		NormalizeColorHistogram(&ch_model, 1, n_images_in_model, &ch_model);
	n_images_in_model = 1;
	CopyModelHistogramToRef();
}

void WriteModel(char *fname)
{
	if (n_images_in_model>0)
		NormalizeColorHistogram(&ch_model, 1, n_images_in_model, &ch_model);
	n_images_in_model = 1;
	WriteColorHistogram(fname, &ch_model);
}

void ReadModel(char *fname)
{
	ReadColorHistogram(fname, &ch_model);
//	NormalizeColorHistogram(&ch_model, 15, 10, &ch_model);
	CopyModelHistogramToRef();
}

void ShowPixelListsAndOutlines()
{
	ShowPixelListsAndOutlines2(&search_strategy, add_lists, sub_lists);
}

void PrintBinContents(unsigned char b,
					  unsigned char g,
					  unsigned char r)
{
	int col1, col2, col3;
	int n_pixels;
	int indx;

	BGRToColorBins(b, g, r, &col1, &col2, &col3);
	indx = col3*N_BINS_COLOR1*N_BINS_COLOR2+col2*N_BINS_COLOR1 + col1;
	assert(indx >= 0 && indx < N_BINS_TOT);
	n_pixels = ch_ref.val[indx];
	color_bin_of_interest = indx;
//	sprintf(msg, "Pixels in reference histogram:  %4d/%4d   current: %4d/%4d        ", 
//		n_pixels, SizeOfColorHistogram(&ch_ref), ch_curr.val[indx], SizeOfColorHistogram(&ch_curr));
//	my_cdc->TextOut(5, 225, msg);
}

void SaveErrorFunctions()
{
	FILE *fpg, *fpc;
	int i;
	int x, y, sz;
	int newx, newy, newsz;
	int goodi;

	fpg = fopen("grad_err.m", "w");
	fpc = fopen("color_err.m", "w");
	fprintf(fpg, "grad_err = [ ...\n");
	fprintf(fpc, "color_err = [ ...\n");
	assert(N_STATE_VARIABLES == 3);
	sz = 0;
	for (y = -msy ; y <= msy ; y++) {
		for (x = -msx ; x <= msx ; x++) {
			for (i=0 ; i<search_strategy.n_states_to_check ;i++) {
				newx  = search_strategy.ds[N_STATE_VARIABLES*i];
				newy  = search_strategy.ds[N_STATE_VARIABLES*i+1];
				newsz = search_strategy.ds[N_STATE_VARIABLES*i+2];
				if (x==newx && y==newy && sz==newsz) {
					goodi = i;
				}
			}
			if (goodi>=0) {
				fprintf(fpg, "%3d ", grad_scores[goodi]);
				fprintf(fpc, "%3d ", color_scores[goodi]);
			}
		}
		fprintf(fpg, "; ...\n");
		fprintf(fpc, "; ...\n");
	}
	fprintf(fpg, "];\n");
	fprintf(fpc, "];\n");
	fclose(fpg);
	fclose(fpc);
}
